// SPDX-License-Identifier: BSD-3-Clause

#include "consumer.h"
#include <pthread.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include "ring_buffer.h"
#include "packet.h"
#include "utils.h"

void consumer_thread(void *args)
{
	/* TODO: implement consumer thread */
	so_packet_t packet;
	so_consumer_ctx_t *ctx = (so_consumer_ctx_t *)args;

	while (true) {
		pthread_mutex_lock(&ctx->producer_rb->mutex);

		while (ctx->producer_rb->len == 0 && ctx->producer_rb->stop == 0)
			pthread_cond_wait(&ctx->producer_rb->not_empty, &ctx->producer_rb->mutex);

		if (ctx->producer_rb->stop == 1 && ctx->producer_rb->len == 0) {
			pthread_mutex_unlock(&ctx->producer_rb->mutex);
			break;
		}

		pthread_mutex_unlock(&ctx->producer_rb->mutex);
		int ok = ring_buffer_dequeue(ctx->producer_rb, &packet, sizeof(so_packet_t));

		if (ok == -1)
			continue;

		pthread_mutex_lock(&ctx->timestamp);
		if (packet.hdr.timestamp >= ctx->lasttimestamp) {
			ctx->lasttimestamp = packet.hdr.timestamp;

			pthread_mutex_lock(&ctx->logs);

			so_action_t action = process_packet(&packet);
			unsigned long hash = packet_hash(&packet);

			fprintf(ctx->f, "%s %016lx %lu\n", RES_TO_STR(action), hash, packet.hdr.timestamp);

			pthread_mutex_unlock(&ctx->logs);
		}
		pthread_mutex_unlock(&ctx->timestamp);
	}
}

int create_consumers(pthread_t *tids,
	int num_consumers,
	struct so_ring_buffer_t *rb,
	const char *out_filename)
{
	FILE *f = fopen(out_filename, "a");

	if (!f) {
		printf("Failed to open output file\n");
		return -1;
	}

	for (int i = 0; i < num_consumers; i++) {
		/*
		 * TODO: Launch consumer threads
		 **/

		so_consumer_ctx_t *ctx = malloc(sizeof(so_consumer_ctx_t));

		if (!ctx) {
			printf("Malloc failed\n");
			fclose(f);
			for (int j = 0; j < i; j++)
				pthread_join(tids[j], NULL);
			return -1;
		}

		ctx->f = f;
		ctx->filename = out_filename;
		ctx->producer_rb = rb;
		ctx->lasttimestamp = 0;
		pthread_mutex_init(&ctx->logs, NULL);
		pthread_mutex_init(&ctx->timestamp, NULL);

		if (pthread_create(&tids[i], NULL, (void *)consumer_thread, ctx) != 0) {
			printf("Error creating thread %d\n", i);
			for (int j = 0; j < i; j++)
				pthread_join(tids[j], NULL);
			pthread_mutex_destroy(&ctx->logs);
			pthread_mutex_destroy(&ctx->timestamp);
			free(ctx);
			fclose(f);
			return -1;
		}
	}
	return num_consumers;
}
